#######################################################################
# Copyright (c) 2021 Josh Luthy BD Life Sciences - Informatics,
# Ashland, OR
#
# Plugin R Script Template Designed and Written by: Ian Taylor and Josh Luthy
#
# CytoNorm algorithm and associated R package developed by Sofie Van Gassen, PhD
#
# License
# The software is distributed under the terms of the
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
#
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct,
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this
# software.
######################################################################

options(warn = - 1)

## INSTALL REQUIRED PACKAGES ##
#######################
# Install from CRAN

## TRY TO AUTO INSTALL ##

# Install function for packages
packages<-function(x){
  x<-as.character(match.call()[[2]])
  if (!require(x,character.only=TRUE)){
    install.packages(pkgs=x,repos="http://cran.r-project.org")
    require(x,character.only=TRUE)
  }
}

# Try to write a special function for devtools installs:
devPackages<-function(x){
  if (!require(x,character.only=TRUE)){
    suppressMessages(suppressWarnings(devtools::install_github(x)))
  }else{
    require(x,character.only=TRUE)
  }
}

fCorePack<-function(x){
  x<-as.character(match.call()[[2]])
  if (!require(x,character.only=TRUE)){
    BiocManager::install(x)
    require(x,character.only=TRUE)
  }
}

packages(devtools)
packages(BiocManager)
packages(dplyr)
fCorePack(FlowSOM)
fCorePack(CytoML)
# devPackages("saeyslab/FlowSOM")
devPackages("saeyslab/CytoNorm")
library("CytoNorm")
# fCorePack(FlowSOM)
packages(ggplot2)
packages(pheatmap)
#######################
###############################
##         LOAD INPUTS       ##
###############################

isWindows <- IS_WINDOWS
outPath <- "FJ_CSV_OUPUT_FILE"

if (isWindows){
  outPath <- gsub("/", "\\\\", outPath)
}

tempDir <- paste0(outPath,"/temp")

if (isWindows){
  tempDir <- gsub("/", "\\\\", tempDir)
}

inCSVPath <- "FJ_DATA_FILE_PATH"
sampleName <- "FJ_SAMPLE_NAME"
applyOn <- "FJ_APPLY_ON"

# TODO: Settings being imported from Java - switches can be repurposed for needs in CytoNorm
timeStamp <- "FJ_OVERLAY_PLOTS"

## Repurposing 'FJ FLIP GRAPH' to the Hashmap of 'samples : control' status
hashList <- list(FJ_FLIP_GRAPH)

nCluster <- FJ_NCLUST

## Repurposing 'FJ TRIM' to ... batches
batchList <- list(FJ_TRIM)
parLegendPosition <- "FJ_LEGEND_POS" ## "left", "top", "right", "bottom"
catParRaw <- "FJ_CAT_PAR"
# createStatTable <- FJ_STATS

## This should be the number of samples (or batches) we're trying to correct
numSamples <- length(hashList)

## Get the DataSet Ready for CytoNorm ##
########################################

outPuts <- paste0(inCSVPath, "/CytoNorm_Dataset_", timeStamp)

# Let's get the FCS file lists addressed the way CytoNorm expects to find them.

data02 <- data.frame( File = character(),
                      Path = character(),
                      Type = character(),
                      Batch = character(),
                      stringsAsFactors = F )

for (i in 1:numSamples) {
  ithHash <- as.character(hashList[i])

  ithBatchHash <- as.character(batchList[i])

  # ithHash <- as.character(hashList[1])
  # tidyselect::vars_select(ithHash, contains(":true"))
  # ":true" %in% ithHash
  # grepl(":true", ithHash)
  # sub(".*/", "", sub(":true","",ithHash))

  if (grepl(":true", ithHash)) {   # }, ignore.case = F, vars = ithHash)) {
    # put these samples into the control group (type = 1)
    data02[i,] <- data.frame( File = sub(".*/", "", sub(":true","",ithHash)),
                              Path = sub(":true","",ithHash),
                              Type = "1",
                              Batch = sub(".*:", "", ithBatchHash),
                              stringsAsFactors = F )


  } else {
    # put these samples inot the test group (type = 2)
    data02[i,] <- data.frame( File = sub(".*/", "", sub(":false","",ithHash)),
                              Path = sub(":false","",ithHash),
                              Type = "2",
                              Batch = sub(".*:", "", ithBatchHash),
                              stringsAsFactors = F )
  }
}

#   ithControl <- gsub(":*", "", ithHash)
#   split(hashList[1], ':', drop = F)
#   splitR <- split(hashList)

data <- data02

########################################
########################################



## Initial Treatment of Dataset ##
##################################

data$Type <- c("1" = "Train", "2" = "Validation")[data$Type]

train_data <- dplyr::filter(data, Type == "Train")
validation_data <- dplyr::filter(data, Type == "Validation")

ff <- flowCore::read.FCS(data$Path[1])

channels <- flowCore::colnames(ff)
transformList <- flowCore::transformList(channels,
                                         CytoNorm::cytofTransform)

transformList.reverse <- flowCore::transformList(channels,
                                                 CytoNorm::cytofTransform.reverse)


##  Step 02 of Initial Workflow ##
##################################

fsom <- CytoNorm::prepareFlowSOM(train_data$Path,
                                 channels,
                                 nCells = 6000,
                                 FlowSOM.params = list(xdim = 5,
                                                       ydim = 5,
                                                       nClus = nCluster,
                                                       scale = FALSE), #TODO place 'colsToUse' here for user selected params
                                 transformList = transformList,
                                 seed = 1)

cVals = sort(c(5,nCluster,10,15))
png("NUL")
pdf("NUL")

cvs <- testCV(fsom, cluster_values = cVals, plot=FALSE)

p <- PlotOverviewCV(fsom, cvs)

ggsave(paste0(outPuts,"/CVcompare.png"), p)

dev.off()

eval(parse(text=paste0("cvs","$","pctgs","$","`",nCluster,"`")))


##  Step 03 of Initial Workflow ##
##################################

tempDir <- paste0(outPuts,"/temp")

if (isWindows){
  tempDir <- gsub("/", "\\\\", tempDir)
} else {
  tempDir <- "./Tmp"
}

model <- CytoNorm.train(files = train_data$Path,
                          labels = train_data$Batch,
                          channels = channels,
                          transformList = transformList,
                          outputDir = tempDir,
                          # plot = FJ_SOMMARY_PLOTS, TODO: implement this option as checkbox (summary plots).
                          FlowSOM.params = list(nCells = 10000,
                                                xdim = 10,
                                                ydim = 10,
                                                nClus = nCluster,
                                                scale = FALSE), #TODO use colsToUse to train on selected params
                          normMethod.train = QuantileNorm.train,
                          normParams = list(nQ = 101,
                                            goal = "mean"),
                          seed = 1,
                          verbose = TRUE)


##  Step 04 of Initial Workflow ##
##################################

## Where are you putting all the outputs? the OutPutDir

OutPutDir <- paste0(outPuts,"/Normalized")

if (isWindows){
  OutPutDir <- gsub("/", "\\\\", OutPutDir)
}else {

}

CytoNorm.normalize(model = model,
                    files = validation_data$Path,
                    labels = validation_data$Batch,

                    transformList = transformList,
                    transformList.reverse = transformList.reverse,
                    normMethod.normalize = QuantileNorm.normalize,
                    outputDir = OutPutDir,
                    prefix = "Norm_",
                    clean = TRUE,
                    verbose = TRUE)


##################################
##################################
### CONFIRM THE PLUGIN HAS FINALIZED ###
########################################
write.table(data.frame(1,2,3,5,8), file=paste0(outPuts,"/fin.tsv"), sep="  ", col.names=F, row.names = F)
########################################
########################################
# R seems to be saving the .RData when exiting, so let's clean up to at least make that tiny (i.e., empty environment)
rm(list=ls())
rm(tempDir)